package config

import (
	"fmt"
	"github.com/mitchellh/mapstructure"
	log "github.com/sirupsen/logrus"
	"github.com/spf13/viper"
	"os"
	"path/filepath"
	"reflect"
	"strings"
)

type Config struct {
	Log   LogConfig   `yaml:"log"`
	Param ParamConfig `yaml:"param"`
	Pair  PairConfig  `yaml:"pair"`
	Rpc   RpcConfig   `yaml:"rpc"`
	Users string      `yaml:"users"`
}

type ParamConfig struct {
	Routine  int   `yaml:"routine"`
	Volume   int64 `yaml:"volume"`
	Count    int   `yaml:"count"`
	GasPrice int64 `yaml:"gas_price"`
	Slippage int64 `yaml:"slippage"`
	Deadline int64 `yaml:"deadline"`
}

type PairConfig struct {
	Pair  string `yaml:"pair"`
	Token string `yaml:"token"`
}

type RpcConfig struct {
	Env      string `yaml:"env"`
	PoolSize int    `yaml:"pool"`
	Url      string `yaml:"url"`
	Proxy    string `yaml:"proxy"`
}

type LogConfig struct {
	Level string `yaml:"level"`
}

var Global *Config

func InitConfig(cfgFile string) error {
	exePath, err := os.Executable()
	if err != nil {
		fmt.Println("Error:", err)
		return err
	}
	log.Warn("execute file is :", exePath)
	exeDir := filepath.Dir(exePath)

	os.Chdir(exeDir)

	setDefaultConfig()

	if cfgFile != "" {
		viper.SetConfigFile(cfgFile)
	} else {
		viper.AddConfigPath(".")
		viper.SetConfigName("config")
		viper.SetConfigType("yaml")
	}

	viper.SetEnvPrefix("SWAPPER")
	viper.SetEnvKeyReplacer(strings.NewReplacer(".", "_"))
	viper.AutomaticEnv()

	if err := viper.ReadInConfig(); err != nil {
		if _, ok := err.(viper.ConfigFileNotFoundError); !ok {
			return fmt.Errorf("failed to read config file: %w", err)
		}
		log.Warn("No config file found, using defaults")
	}

	decoderConfig := &mapstructure.DecoderConfig{
		Result:           &Global,
		TagName:          "yaml",
		WeaklyTypedInput: true,
	}

	decoder, err := mapstructure.NewDecoder(decoderConfig)
	if err != nil {
		return fmt.Errorf("failed to create decoder: %w", err)
	}

	if err := decoder.Decode(viper.AllSettings()); err != nil {
		return fmt.Errorf("failed to decode config: %w", err)
	}

	if err := Global.Validate(); err != nil {
		return fmt.Errorf("config validation failed: %w", err)
	}

	return nil
}

// setDefaultConfig
func setDefaultConfig() {
	defaultValues := structToMap(DefaultConfig)
	for key, value := range defaultValues {
		viper.SetDefault(key, value)
	}
}

func structToMap(obj interface{}) map[string]interface{} {
	result := make(map[string]interface{})
	fillMap(reflect.ValueOf(obj), "", result)
	return result
}

func fillMap(v reflect.Value, prefix string, result map[string]interface{}) {
	t := v.Type()
	for i := 0; i < t.NumField(); i++ {
		field := t.Field(i)
		value := v.Field(i)

		tag := field.Tag.Get("yaml")
		if tag == "" {
			tag = strings.ToLower(field.Name)
		}
		key := tag
		if prefix != "" {
			key = prefix + "." + tag
		}

		if value.Kind() == reflect.Struct {
			fillMap(value, key, result)
		} else {
			result[key] = value.Interface()
		}
	}
}

func (c *Config) Validate() error {
	return nil
}

func (c *Config) GetRootDir() string {
	return "./"
}
