package work

import (
	"code.wuban.net.cn/service/pancakeswapper/clientpool"
	"code.wuban.net.cn/service/pancakeswapper/config"
	"code.wuban.net.cn/service/pancakeswapper/contracts/buildparam"
	"code.wuban.net.cn/service/pancakeswapper/contracts/erc20"
	router2 "code.wuban.net.cn/service/pancakeswapper/contracts/router"
	"code.wuban.net.cn/service/pancakeswapper/contracts/v3router"
	"code.wuban.net.cn/service/pancakeswapper/pancake"
	"code.wuban.net.cn/service/pancakeswapper/types"
	"context"
	"encoding/csv"
	"encoding/hex"
	"encoding/json"
	"fmt"
	"github.com/ethereum/go-ethereum/accounts/abi"
	"github.com/ethereum/go-ethereum/accounts/abi/bind"
	"github.com/ethereum/go-ethereum/common"
	ethtypes "github.com/ethereum/go-ethereum/core/types"
	"github.com/ethereum/go-ethereum/crypto"
	"github.com/ethereum/go-ethereum/ethclient"
	log "github.com/sirupsen/logrus"
	"math/big"
	"os"
	"strings"
	"time"
)

var (
	failedHistory = make(chan *types.Account, 1000000)
	dec           = new(big.Int).SetUint64(1000000000000000000)
	fee           = big.NewInt(10000)
)

func LoadAccounts(file string) []*types.Account {
	data, err := os.ReadFile(file)
	if err != nil {
		panic(fmt.Errorf("read accounts file failed: %v", err))
	}
	accs := make([]*types.Account, 0)
	err = json.Unmarshal(data, &accs)
	if err != nil {
		panic(fmt.Errorf("unmarshal accounts to arrry failed: %v", err))
	}
	return accs
}

func LoadAccountsFromCsv(filepath string) []*types.Account {
	// parse csv file, column is address, private key, recipient.
	file, err := os.Open(filepath)
	if err != nil {
		panic(fmt.Errorf("Error opening file:", err))
	}
	defer file.Close()

	// Create a CSV reader
	reader := csv.NewReader(file)
	records, err := reader.ReadAll()
	if err != nil {
		panic(fmt.Errorf("Error reading CSV file:", err))
	}
	var parsedRecords []*types.Account
	for i, record := range records {
		// Ensure the record has the expected number of columns
		if len(record) != 3 {
			log.Errorf("Skipping malformed record at row %d: %+v\n", i, record)
			continue
		}
		if i == 0 && !strings.HasPrefix(record[0], "0x") {
			log.Warningf("Skipping malformed record at row %d: %+v\n", i, record)
			// skip title line.
			continue
		}

		// Map the fields to the Record struct
		parsedRecord := &types.Account{
			Address:    strings.TrimSpace(record[0]),
			PrivateKey: strings.TrimSpace(record[1]),
			Recipient:  strings.TrimSpace(record[2]),
		}

		parsedRecords = append(parsedRecords, parsedRecord)
	}

	log.WithField("records", len(parsedRecords)).Info("Successfully loaded accounts from csv file")
	return parsedRecords
}

func addTofailed(acc *types.Account) {
	IncrFailedUser()
	failedHistory <- acc
}

func waitTx(client *ethclient.Client, tx *ethtypes.Transaction) *ethtypes.Receipt {
	tm := time.NewTicker(time.Second)
	defer tm.Stop()
	for {
		select {
		case <-tm.C:
			r, _ := client.TransactionReceipt(context.TODO(), tx.Hash())
			if r != nil {
				return r
			}
		}
	}
}

func accountApproveWithPermit(client *ethclient.Client, acc *types.Account, env types.Env) error {
	cfg := config.Global
	permit2 := common.HexToAddress(env.Permit2)
	usdt, err := erc20.NewTokenContract(common.HexToAddress(env.USDT), client)
	if err != nil {
		return fmt.Errorf("erc20.NewTokenContract failed: %v", err)
	}
	// router need set to Universal Router.
	router, err := buildparam.NewParamContract(common.HexToAddress(env.UniversalRouter), client)
	if err != nil {
		return fmt.Errorf("build param contract failed: %v", err)
	}

	txopts := &bind.TransactOpts{
		From:     acc.Addr,
		Signer:   acc.SignTx,
		GasPrice: new(big.Int).Mul(big.NewInt(cfg.Param.GasPrice), big.NewInt(1e9)),
	}
	addNewTx := func(tx *ethtypes.Transaction) {
		waitTx(client, tx)
	}
	allow, err := usdt.Allowance(&bind.CallOpts{
		From:        acc.Addr,
		BlockNumber: nil,
		Context:     context.Background(),
	}, acc.Addr, permit2)
	if allow.Int64() == 0 {
		if tx, err := usdt.Approve(txopts, permit2, abi.MaxUint256); err != nil {
			return fmt.Errorf("usdt approve permit2 failed: %v", err)
		} else {
			addNewTx(tx)
		}
	}

	permitData, err := pancake.MakePermit2PermitData2(acc, env, client)
	if err != nil {
		return fmt.Errorf("make permit data: %v", err)
	}

	inputs := make([][]byte, 0)
	inputs = append(inputs, permitData)
	fmt.Println("permit input data is ", hex.EncodeToString(permitData))
	deadline := time.Now().Unix() + cfg.Param.Deadline*60
	routerAbi, _ := abi.JSON(strings.NewReader(router2.RouterMetaData.ABI))
	txdata, err := routerAbi.Pack("execute0", []byte{byte(pancake.PERMIT2_PERMIT)}, inputs, big.NewInt(deadline))
	if err != nil {
		fmt.Println()
		return fmt.Errorf("pack execute failed: %v", err)
	}
	fmt.Println("txdata is ", hex.EncodeToString(txdata))

	// set permit with universal router execute.

	if execTx, err := router.Execute(txopts, []byte{byte(pancake.PERMIT2_PERMIT)}, inputs, big.NewInt(deadline)); err != nil {
		return fmt.Errorf("execute permit2 permit failed: %v", err)
	} else {
		addNewTx(execTx)
	}
	return nil
}

func accountApprove(client *ethclient.Client, acc *types.Account, env types.Env) error {
	cfg := config.Global
	usdt, err := erc20.NewTokenContract(common.HexToAddress(env.USDT), client)
	if err != nil {
		return fmt.Errorf("erc20.NewTokenContract failed: %v", err)
	}

	txopts := &bind.TransactOpts{
		From:     acc.Addr,
		Signer:   acc.SignTx,
		GasPrice: new(big.Int).Mul(big.NewInt(cfg.Param.GasPrice), big.NewInt(1e9)),
	}

	addNewTx := func(tx *ethtypes.Transaction) {
		waitTx(client, tx)
	}
	allow, err := usdt.Allowance(&bind.CallOpts{
		From:        acc.Addr,
		BlockNumber: nil,
		Context:     context.Background(),
	}, acc.Addr, common.HexToAddress(env.Router))
	if err != nil {
		log.WithError(err).Warn("allow permit2 permit failed")
		return fmt.Errorf("get usdt allowance failed: %v", err)
	}
	if allow == nil || allow.Int64() == 0 {
		if tx, err := usdt.Approve(txopts, common.HexToAddress(env.Router), abi.MaxUint256); err != nil {
			return fmt.Errorf("usdt approve pool failed: %v", err)
		} else {
			addNewTx(tx)
		}
	}
	return nil
}

func HandlerAccountSwap(acc *types.Account, env types.Env, idx uint) {
	cfg := config.Global
	client := clientpool.GetClient()
	pkstr := acc.PrivateKey
	if strings.HasPrefix(pkstr, "0x") || strings.HasPrefix(pkstr, "0X") {
		pkstr = acc.PrivateKey[2:]
	}
	var err error
	acc.PK, err = crypto.HexToECDSA(pkstr)
	if err != nil {
		addTofailed(acc)
		WriteLog(fmt.Sprintf("[%s] failed: invalid private hex string: %v", acc.Address, pkstr))
		return
	}
	acc.Addr = crypto.PubkeyToAddress(acc.PK.PublicKey)

	if strings.ToLower(acc.Addr.Hex()) != strings.ToLower(acc.Address) {
		addTofailed(acc)
		WriteLog(fmt.Sprintf("[%s] failed: given address not match with private key (%s), please check", acc.Address, acc.Addr))
		return
	}

	if err := accountApprove(client, acc, env); err != nil {
		addTofailed(acc)
		WriteLog(fmt.Sprintf("[%s] do approve failed: %v", acc.Address, err))
		return
	}

	txopts := &bind.TransactOpts{
		From:     acc.Addr,
		Signer:   acc.SignTx,
		GasPrice: new(big.Int).Mul(big.NewInt(cfg.Param.GasPrice), big.NewInt(1e9)),
	}
	addNewTx := func(tx *ethtypes.Transaction) {
		IncrTotalTx()
		receipt := waitTx(client, tx)
		for _, log := range receipt.Logs {
			if log.Topics[0] == pancake.TransferTopic {
				token, _ := erc20.NewTokenContract(common.HexToAddress(cfg.Pair.Token), client)
				event, _ := token.ParseTransfer(*log)
				if event.To == acc.Addr || event.To == common.HexToAddress(acc.Recipient) {
					val, _ := new(big.Float).SetInt(event.Value).Float64()
					val = val / 1e18
					WriteLog(fmt.Sprintf("[%s] receive %f token", acc.Address, val))
					IncrReceiveToken(val)
				}
				if event.From == acc.Addr {
					val, _ := new(big.Float).SetInt(event.Value).Float64()
					val = val / 1e18
					WriteLog(fmt.Sprintf("[%s] cost %f token", acc.Address, val))
					IncrCostToken(val)
				}
			}
		}
		IncrFinishedTx()
	}

	for i := 0; i < cfg.Param.Count; i++ {
		in := new(big.Int).Mul(big.NewInt(cfg.Param.Volume), dec)
		in = new(big.Int).Div(in, big.NewInt(10)) // param volume 1 is 0.1
		v3r, err := v3router.NewV3Router(common.HexToAddress(env.Router), client)

		// calc amountOutMin with price and slippage.
		amountOutMin, err := pancake.GetAmountOutMin(cfg, client, env, in, cfg.Param.Slippage, fee)
		if err != nil {
			addTofailed(acc)
			WriteLog(fmt.Sprintf("[%s] failed: failed to get amount out min: %v", acc.Address, err))
			return
		}

		deadline := time.Now().Unix() + cfg.Param.Deadline*60
		if execTx, err := v3r.ExactInputSingle(txopts, v3router.ISwapRouterExactInputSingleParams{
			TokenIn:           common.HexToAddress(env.USDT),
			TokenOut:          common.HexToAddress(cfg.Pair.Token),
			Recipient:         common.HexToAddress(acc.Recipient),
			AmountIn:          in,
			AmountOutMinimum:  amountOutMin,
			Deadline:          big.NewInt(deadline),
			Fee:               fee,
			SqrtPriceLimitX96: big.NewInt(0),
		}); err != nil {
			addTofailed(acc)
			WriteLog(fmt.Sprintf("[%s] failed: failed execute swap: %v", acc.Address, err))
			return
		} else {
			addNewTx(execTx)
		}
	}
	WriteLog(fmt.Sprintf("[%s] finish: have finished all swap", acc.Address))
	IncrFinishedUser()
}

func HandlerAccountSwapWithUniversalRouter(acc *types.Account, env types.Env, idx uint) {
	cfg := config.Global
	client := clientpool.GetClient()
	pkstr := acc.PrivateKey
	if strings.HasPrefix(pkstr, "0x") || strings.HasPrefix(pkstr, "0X") {
		pkstr = acc.PrivateKey[2:]
	}

	var err error
	acc.PK, err = crypto.HexToECDSA(pkstr)
	if err != nil {
		addTofailed(acc)
		WriteLog(fmt.Sprintf("[%s] failed: invalid private hex string: %v", acc.Address, pkstr))
		return
	}
	acc.Addr = crypto.PubkeyToAddress(acc.PK.PublicKey)

	if strings.ToLower(acc.Addr.Hex()) != strings.ToLower(acc.Address) {
		addTofailed(acc)
		WriteLog(fmt.Sprintf("[%s] failed: given address not match with private key (%s), please check", acc.Address, acc.Addr))
		return
	}

	if err := accountApproveWithPermit(client, acc, env); err != nil {
		addTofailed(acc)
		WriteLog(fmt.Sprintf("[%s] do approve failed: %v", acc.Address, err))
		return
	}

	router, err := buildparam.NewParamContract(common.HexToAddress(env.UniversalRouter), client)
	if err != nil {
		addTofailed(acc)
		WriteLog(fmt.Sprintf("[%s] failed: failed to create router contract: %v", acc.Address, err))
		return
	}

	txopts := &bind.TransactOpts{
		From:     acc.Addr,
		Signer:   acc.SignTx,
		GasPrice: new(big.Int).Mul(big.NewInt(cfg.Param.GasPrice), big.NewInt(1e9)),
	}

	addNewTx := func(tx *ethtypes.Transaction) {
		IncrTotalTx()
		receipt := waitTx(client, tx)
		for _, log := range receipt.Logs {
			if log.Topics[0] == pancake.TransferTopic {
				token, _ := erc20.NewTokenContract(common.HexToAddress(cfg.Pair.Token), client)
				event, _ := token.ParseTransfer(*log)
				if event.To == acc.Addr {
					val, _ := new(big.Float).SetInt(event.Value).Float64()
					val = val / 1e18
					WriteLog(fmt.Sprintf("[%s] receive %f token", acc.Address, val))
					IncrReceiveToken(val)
				}
				if event.From == acc.Addr {
					val, _ := new(big.Float).SetInt(event.Value).Float64()
					val = val / 1e18
					IncrCostToken(val)
				}
			}
		}
		IncrFinishedTx()
	}

	for i := 0; i < cfg.Param.Count; i++ {
		in := new(big.Int).Mul(big.NewInt(cfg.Param.Volume), dec)
		in = new(big.Int).Div(in, big.NewInt(10)) // param volume 1 is 0.1

		// calc amountOutMin with price and slippage.
		amountOutMin, err := pancake.GetAmountOutMin(cfg, client, env, in, cfg.Param.Slippage, fee)
		if err != nil {
			addTofailed(acc)
			WriteLog(fmt.Sprintf("[%s] failed: failed to get amount out min: %v", acc.Address, err))
			return
		}
		param := pancake.ParamV3SwapExactIn{
			Recipient:    common.HexToAddress(acc.Recipient),
			AmountIn:     in,
			AmountOutMin: amountOutMin,
			Path:         nil,
			PayerIsUser:  true,
		}
		data, err := pancake.MakeV3SwapExactInData(env, param, cfg)
		if err != nil {
			addTofailed(acc)
			WriteLog(fmt.Sprintf("[%s] failed: failed to make swap data: %v", acc.Address, err))
			return
		}
		inputs := make([][]byte, 0)
		inputs = append(inputs, data)
		deadline := time.Now().Unix() + cfg.Param.Deadline*60
		if execTx, err := router.Execute(txopts, []byte{byte(pancake.V3_SWAP_EXACT_IN)}, inputs, big.NewInt(deadline)); err != nil {
			addTofailed(acc)
			WriteLog(fmt.Sprintf("[%s] failed: failed execute swap: %v", acc.Address, err))
			return
		} else {
			addNewTx(execTx)
		}
	}
	WriteLog(fmt.Sprintf("[%s] finish: have finished all swap", acc.Address))
	IncrFinishedUser()
}

func Finish() int {
	close(failedHistory)
	// read all from failedHistory
	var failed []*types.Account
	for acc := range failedHistory {
		failed = append(failed, acc)
	}
	if len(failed) > 0 {
		data, _ := json.MarshalIndent(failed, "", "  ")
		os.WriteFile("failed.json", data, 0666)
	}

	WriteFinished()
	return len(failed)
}
